#include "unittest.h"
#include "../resources.h"
#include "../../shared/fileHandling.h"
#include "../multithreading.h"
#include "../../algorithm.h"
#include "../../shared/globalFunctions.h"
#include "../../ui/util.h"
#include "../filter.h"

using namespace FreeFileSync;



TEST(Miscellaneous, RemoveRowsFromVector)
{
    std::vector<int> test; //0 1 2 3 4 5 6
    for (int i = 0; i < 7; ++i)
        test.push_back(i);

    std::set<unsigned int> rowsToRemove;
    rowsToRemove.insert(-41); //invalid
    rowsToRemove.insert(1);
    rowsToRemove.insert(2);
    rowsToRemove.insert(5);
    rowsToRemove.insert(15);  //invalid
    rowsToRemove.insert(100); //invalid

    globalFunctions::removeRowsFromVector(rowsToRemove, test);
    //0 3 4 6

    ASSERT_EQ(unsigned(4), test.size());
    ASSERT_EQ(0, test[0]);
    ASSERT_EQ(3, test[1]);
    ASSERT_EQ(4, test[2]);
    ASSERT_EQ(6, test[3]);
}


TEST(Formatting, DirectoryNames)
{
    ASSERT_PRED_FORMAT2(assertEqual, FreeFileSync::getFormattedDirectoryName(wxT("  c: ")).c_str(), wxT("c:\\"));
    ASSERT_PRED_FORMAT2(assertEqual, FreeFileSync::getFormattedDirectoryName(wxT("  c:\\test\\ ")).c_str(), wxT("c:\\test\\"));
    ASSERT_PRED_FORMAT2(assertEqual, FreeFileSync::getFormattedDirectoryName(wxT("  \\abc\\ ")).c_str(), wxT("\\abc\\"));
    ASSERT_PRED_FORMAT2(assertEqual, FreeFileSync::getFormattedDirectoryName(wxT("  \\a\\b\\c ")).c_str(), wxT("\\a\\b\\c\\"));
    ASSERT_PRED_FORMAT2(assertEqual, FreeFileSync::getFormattedDirectoryName(wxT("  \\a\\b\\c d  ")).c_str(), wxT("\\a\\b\\c d\\"));
    ASSERT_PRED_FORMAT2(assertEqual, FreeFileSync::getFormattedDirectoryName(wxT("  \\a b/b\\c d  ")).c_str(), wxT("\\a b/b\\c d\\"));
}


//predicate-formatter for asserting that two vector<Zstring> are equal
testing::AssertionResult assertEqual(const char* a_expr, const char* b_expr, const std::vector<Zstring>& a, const std::vector<Zstring>& b)
{
    if (a == b)
        return testing::AssertionSuccess();

    testing::Message msg;
    msg << a_expr << " = " << '\n';
    for (std::vector<Zstring>::const_iterator i = a.begin(); i != a.end(); ++i)
        msg << i->c_str() << '\n';

    msg << " and " << b_expr << " = " << '\n';
    for (std::vector<Zstring>::const_iterator i = b.begin(); i != b.end(); ++i)
        msg << i->c_str() << '\n';

    msg << " not equal!";

    return testing::AssertionFailure(msg);
}

std::vector<Zstring> compoundStringToFilter(const Zstring& filterString);

TEST(Formatting, FilterString)
{
    std::vector<Zstring> filterActual = compoundStringToFilter(wxT("   A\n   b;; c;\n\n;\n d    ;E    f  \n;& ;g\\"));
    std::vector<Zstring> filterExpected;
    filterExpected.push_back(wxT("A"));
    filterExpected.push_back(wxT("b"));
    filterExpected.push_back(wxT("c"));
    filterExpected.push_back(wxT("d"));
    filterExpected.push_back(wxT("E    f"));
    filterExpected.push_back(wxT("&"));
    filterExpected.push_back(wxT("g\\"));
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);

    filterActual = compoundStringToFilter(wxT("   ; ;\n a"));
    filterExpected.clear();
    filterExpected.push_back(wxT("a"));
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);

    filterActual = compoundStringToFilter(wxT("   ; ;\n       ;\n\n"));
    filterExpected.clear();
    ASSERT_PRED_FORMAT2(assertEqual, filterActual, filterExpected);
}


TEST(Formatting, FileSize)
{
    wxString sizeActual = FreeFileSync::formatFilesizeToShortString(807.0);
    wxString sizeExpected = wxString(wxT("807")) + _(" Byte");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(315621376.0);
    sizeExpected = wxString::Format(wxT("%i"), 301) + _(" MB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(48954.0);
    sizeExpected = wxString(wxT("47.8")) + _(" kB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(35032904.0);
    sizeExpected = wxString(wxT("33.4")) + _(" MB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(3084582912.0);
    sizeExpected = wxString(wxT("2.87")) + _(" GB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);

    sizeActual = FreeFileSync::formatFilesizeToShortString(1001.0);
    sizeExpected = wxString(wxT("0.97")) + _(" kB");
    ASSERT_PRED_FORMAT2(assertEqual, sizeActual, sizeExpected);
}


TEST(Filtering, Standard)
{
    FreeFileSync::FilterProcess filterIncl(wxT("*abc?5*2  ;    abc?3;   name\\; test\\subdir "), wxT(""));

    ASSERT_TRUE(filterIncl.passFileFilter( wxT("zzzzabcz5zzzzz2")));
    ASSERT_FALSE(filterIncl.passFileFilter(wxT("zzzzabczz5zzzz2")));
    ASSERT_FALSE(filterIncl.passFileFilter(wxT("zzzzabcz5zzzzz")));

    ASSERT_TRUE(filterIncl.passFileFilter(wxT("abcz3")));
    ASSERT_FALSE(filterIncl.passFileFilter(wxT("abcz")));

    bool continueTrav = true;
    ASSERT_TRUE(filterIncl.passDirFilter(wxT("name"), &continueTrav));

    ASSERT_FALSE(filterIncl.passDirFilter(wxT("test"), &continueTrav));
    ASSERT_TRUE(continueTrav); //but continue traversing, because subdir might match filter
//------------------------------------------------------------------------------------------

    FreeFileSync::FilterProcess filterExcl(wxT("*"), wxT("asdf?5 ;;; name\\*"));

    ASSERT_FALSE(filterExcl.passFileFilter(wxT("asdfz5")));
    ASSERT_TRUE(filterExcl.passFileFilter(wxT("asdfz")));

    ASSERT_FALSE(filterExcl.passDirFilter(wxT("name"), &continueTrav));
    ASSERT_FALSE(continueTrav);
}


//handle execution of a method while updating the UI
class UpdateDummy : public UpdateWhileExecuting
{
private:
    void longRunner() {} //virtual impl.
};


TEST(Miscellaneous, Standard)
{
    //Windows Recycle Bin should always be accessible
#ifdef FFS_WIN
    ASSERT_TRUE(FreeFileSync::recycleBinExists());
#endif

    //test if instantiation and deletion of multithreading class works correctly (without hang)
    for (int i = 0; i < 50; ++i)
    {
        UpdateDummy d;
    }

    //test if multiple executions work (without hang)
    UpdateDummy d;
    DummyStatusUpdater dummy;
    for (int i = 0; i < 1000; ++i)
    {
        d.waitUntilReady();
        d.execute(&dummy);
    }
}


TEST(Zstring, AfterBefore)
{
    ASSERT_EQ(Zstring(wxT("a/dsfsadf/abc")).AfterLast(wxT('/')), Zstring(wxT("abc")));
    ASSERT_EQ(Zstring(wxT("a/dsfsadf/abc")).BeforeLast(wxT('/')), Zstring(wxT("a/dsfsadf")));
    ASSERT_EQ(Zstring(wxT("a/dsfsadf/abc")).AfterFirst(wxT('/')), Zstring(wxT("dsfsadf/abc")));
    ASSERT_EQ(Zstring(wxT("a/dsfsadf/abc")).BeforeFirst(wxT('/')), Zstring(wxT("a")));
}




